import datatable as dt
import logging
import numpy as np
import pysam
import tqdm
import sys
import random
from bam_filter.utils import (
    create_empty_output_files,
    sort_keys_by_approx_weight,
    is_debug,
    get_arguments,
    check_tmp_dir_exists,
    handle_warning,
    create_output_files,
)
from multiprocessing import Pool, Manager, cpu_count
from functools import partial
import numpy.lib.recfunctions as rf
import gc
from collections import defaultdict
import os
import concurrent.futures
import math
import warnings
from bam_filter.sam_utils import check_bam_file

# import cProfile as prof
# import pstats

log = logging.getLogger("my_logger")


def initialize_subject_weights(data):
    if data.shape[0] > 0:
        # Add a new column for inverse sequence length
        data["s_W"] = 1 / data["slen"]

        # Calculate the sum of weights for each unique source
        sum_weights = np.zeros(int(np.max(data["source"])) + 1)
        np.add.at(sum_weights, data["source"], data["var"])

        # Calculate the normalized weights based on the sum for each query
        query_sum_var = sum_weights[data["source"]]
        data["prob"] = data["var"] / query_sum_var

        return data
    else:
        return None


def resolve_multimaps(data, scale=0.9, iters=10):
    current_iter = 0
    while True:
        progress_bar = tqdm.tqdm(
            total=9,
            desc=f"Iter {current_iter + 1}",
            unit="step",
            disable=False,  # Replace with your logic or a boolean value
            leave=False,
            ncols=80,
        )
        log.debug(f"::: Iter: {current_iter + 1} - Getting scores")
        progress_bar.update(1)
        n_alns = data.shape[0]
        log.debug(f"::: Iter: {current_iter + 1} - Total alignment: {n_alns:,}")

        # Calculate the weights for each subject
        log.debug(f"::: Iter: {current_iter + 1} - Calculating weights...")
        progress_bar.update(1)
        subject_weights = np.zeros(int(np.max(data["subject"])) + 1)
        np.add.at(subject_weights, data["subject"], data["prob"])
        data["s_W"] = subject_weights[data["subject"]] / data["slen"]
        subject_weights = None

        log.debug(f"::: Iter: {current_iter + 1} - Calculating probabilities")
        progress_bar.update(1)
        # Calculate the alignment probabilities
        new_prob = data["prob"] * data["s_W"]
        log.debug("Calculating sum of probabilities")
        progress_bar.update(1)
        prob_sum = data["prob"] * data["s_W"]
        prob_sum_array = np.zeros(int(np.max(data["source"])) + 1)
        np.add.at(prob_sum_array, data["source"], prob_sum)
        prob_sum = None

        # data["prob_sum"] = prob_sum_array[data["source"]]
        data["prob"] = new_prob / prob_sum_array[data["source"]]
        prob_sum_array = None

        log.debug("Calculating query counts")
        progress_bar.update(1)
        # Calculate how many alignments are in each query
        # query_counts = np.zeros(int(np.max(data["source"])) + 1)
        # np.add.at(query_counts, data["source"], 1)

        query_counts = np.bincount(data["source"])

        log.debug("Calculating query counts array")
        progress_bar.update(1)
        # Use a separate array for query counts
        query_counts_array = np.zeros(int(np.max(data["source"])) + 1)
        np.add.at(
            query_counts_array,
            data["source"],
            query_counts[data["source"]],
        )

        log.debug(
            f"::: Iter: {current_iter + 1} - Calculating number of alignments per query"
        )
        progress_bar.update(1)
        data["n_aln"] = query_counts_array[data["source"]]

        log.debug("Calculating unique alignments")
        data["n_aln"] = query_counts_array[data["source"]]
        data_unique = data[data["n_aln"] == 1]
        n_unique = data_unique.shape[0]

        if n_unique == data.shape[0]:
            progress_bar.close()
            log.info("::: ::: No more multimapping reads. Early stopping.")
            return data
        data = data[(data["n_aln"] > 1) & (data["prob"] > 0)]

        # total_n_unique = np.sum(query_counts_array[data["source"]] <= 1)

        query_counts = None
        query_counts_array = None

        log.debug("Calculating max_prob")
        # Keep the ones that have a probability higher than the maximum scaled probability
        max_prob = np.zeros(int(np.max(data["source"])) + 1)
        np.maximum.at(max_prob, data["source"], data["prob"])

        data["max_prob"] = max_prob[data["source"]]
        data["max_prob"] = data["max_prob"] * scale
        # data["max_prob"] = max_prob[data["source"]]
        log.debug(
            f"::: Iter: {current_iter + 1} - Removing alignments with lower probability"
        )
        progress_bar.update(1)
        to_remove = np.sum(data["prob"] < data["max_prob"])

        data = data[data["prob"] >= data["max_prob"]]
        max_prob = None

        # Update the iteration count in the function call
        current_iter += 1
        data["iter"] = current_iter
        data_unique["iter"] = current_iter

        query_counts = np.bincount(data["source"])
        total_n_unique = np.sum(query_counts[data["source"]] <= 1)

        # data_unique["iter"] = current_iter

        # data = np.concatenate([data, data_unique])
        data = np.concatenate([data, data_unique])
        data_unique = None

        keep_processing = to_remove != 0
        log.debug(f"::: Iter: {current_iter} - Removed {to_remove:,} alignments")
        log.debug(f"::: Iter: {current_iter} - Total mapping queries: {n_unique:,}")
        log.debug(
            f"::: Iter: {current_iter} - New unique mapping queries: {total_n_unique:,}"
        )
        log.debug(f"::: Iter: {current_iter} - Alns left: {data.shape[0]:,}")
        progress_bar.update(1)
        progress_bar.close()
        log.info(
            f"::: Iter: {current_iter} - R: {to_remove:,} | U: {total_n_unique:,} | NU: {n_unique:,} | L: {data.shape[0]:,}"
        )
        log.debug(f"::: Iter: {current_iter} - done!")

        if iters > 0 and current_iter >= iters:
            log.info("::: ::: Reached maximum iterations. Stopping.")
            break
        elif not keep_processing:
            log.info("::: ::: No more alignments to remove. Stopping.")
            break
    return data


# def write_reassigned_bam(
#     bam, out_files, threads, entries, sort_memory="1G", min_read_ani=90
# ):
#     if out_files["bam_reassigned"] is not None:
#         out_bam = out_files["bam_reassigned"]
#     else:
#         out_bam = out_files["bam_reassigned_sorted"]

#     samfile = pysam.AlignmentFile(bam, "rb", threads=threads)
#     references = list(entries.keys())
#     refs_dict = {x: samfile.get_reference_length(x) for x in list(entries.keys())}

#     (ref_names, ref_lengths) = zip(*refs_dict.items())

#     refs_idx = {sys.intern(str(x)): i for i, x in enumerate(ref_names)}
#     if threads > 4:
#         write_threads = 4
#     else:
#         write_threads = threads

#     out_bam_file = pysam.AlignmentFile(
#         out_files["bam_reassigned_tmp"],
#         "wb",
#         referencenames=list(ref_names),
#         referencelengths=list(ref_lengths),
#         threads=write_threads,
#     )

#     for reference in tqdm.tqdm(
#         references,
#         total=len(references),
#         leave=False,
#         ncols=80,
#         desc="References processed",
#     ):
#         r_ids = entries[reference]
#         for aln in samfile.fetch(
#             reference=reference, multiple_iterators=False, until_eof=True
#         ):
#             # ani_read = (1 - ((aln.get_tag("NM") / aln.infer_query_length()))) * 100
#             if (aln.query_name, reference) in r_ids:
#                 aln.reference_id = refs_idx[aln.reference_name]
#                 out_bam_file.write(aln)
#     out_bam_file.close()


# def write_to_file(alns, out_bam_file):
#     for aln in tqdm.tqdm(alns, total=len(alns), leave=False, ncols=80, desc="Writing"):
#         out_bam_file.write(aln)


def write_to_file(alns, out_bam_file, header=None):
    for aln in alns:
        out_bam_file.write(pysam.AlignedSegment.fromstring(aln, header))


def process_references_batch(references, entries, bam, refs_idx, threads=4):
    alns = []
    with pysam.AlignmentFile(bam, "rb", threads=threads) as samfile:
        for reference in references:
            r_ids = entries[reference]
            for aln in samfile.fetch(
                reference=reference, multiple_iterators=False, until_eof=True
            ):
                if (aln.query_name, reference) in r_ids:
                    aln.reference_id = refs_idx[aln.reference_name]
                    alns.append(aln.to_string())

    return alns


def write_reassigned_bam(
    bam,
    ref_counts,
    out_files,
    threads,
    entries,
    sort_memory="1G",
    sort_by_name=False,
    min_read_ani=90,
    min_read_length=30,
):
    # if out_files["bam_reassigned"] is not None:
    #     out_bam = out_files["bam_reassigned"]
    # else:
    #     out_bam = out_files["bam_reassigned_sorted"]
    out_bam = out_files["bam_reassigned"]
    samfile = pysam.AlignmentFile(bam, "rb", threads=threads)
    references = list(entries.keys())
    refs_dict = {x: samfile.get_reference_length(x) for x in references}
    header = samfile.header
    samfile.close()
    (ref_names, ref_lengths) = zip(*refs_dict.items())

    refs_idx = {sys.intern(str(x)): i for i, x in enumerate(ref_names)}
    if threads > 4:
        write_threads = 4
    else:
        write_threads = threads

    out_bam_file = pysam.AlignmentFile(
        out_files["bam_reassigned_tmp"],
        "wb",
        referencenames=list(ref_names),
        referencelengths=list(ref_lengths),
        threads=write_threads,
    )

    num_cores = min(threads, cpu_count())
    # batch_size = len(references) // num_cores + 1  # Ensure non-zero batch size
    # batch_size = calc_chunksize(n_workers=num_cores, len_iterable=len(references))
    log.info("::: Creating reference chunks with uniform read amounts...")

    ref_chunks = sort_keys_by_approx_weight(
        input_dict=ref_counts, scale=1, num_cores=num_cores
    )
    num_cores = min(num_cores, len(ref_chunks))
    log.info(f"::: Using {num_cores} cores to write {len(ref_chunks)} chunk(s)")

    with Manager() as manager:
        # Use Manager to create a read-only proxy for the dictionary
        entries = manager.dict(dict(entries))

        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cores) as executor:
            # Use ProcessPoolExecutor to parallelize the processing of references in batches
            futures = []
            for batch_references in tqdm.tqdm(
                ref_chunks,
                total=len(ref_chunks),
                desc="Submitted batches",
                unit="batch",
                leave=False,
                ncols=80,
                disable=is_debug(),
            ):
                future = executor.submit(
                    process_references_batch, batch_references, entries, bam, refs_idx
                )
                futures.append(future)  # Store the future

            # Use a while loop to continuously check for completed futures
            log.info("::: Collecting batches...")

            completion_progress_bar = tqdm.tqdm(
                total=len(futures),
                desc="Completed",
                unit="batch",
                leave=False,
                ncols=80,
                disable=is_debug(),
            )
            completed_count = 0

            # Use as_completed to iterate over completed futures as they become available
            for completed_future in concurrent.futures.as_completed(futures):
                alns = completed_future.result()
                write_to_file(alns=alns, out_bam_file=out_bam_file, header=header)

                # Update the progress bar for each completed write
                completion_progress_bar.update(1)
                completed_count += 1
                completed_future.cancel()  # Cancel the future to free memory
                gc.collect()  # Force garbage collection

            completion_progress_bar.close()
    out_bam_file.close()
    entries = None
    gc.collect()
    # prof.disable()
    # # print profiling output
    # stats = pstats.Stats(prof).strip_dirs().sort_stats("tottime")
    # stats.print_stats(5)  # top 10 rows
    log.info("::: ::: Sorting BAM file...")
    if sort_by_name:
        log.info("::: ::: Sorting by name...")
        pysam.sort(
            "-n",
            "-@",
            str(threads),
            "-m",
            str(sort_memory),
            "-o",
            out_bam,
            out_files["bam_reassigned_tmp"],
        )
    else:
        pysam.sort(
            "-@",
            str(threads),
            "-m",
            str(sort_memory),
            "-o",
            out_bam,
            out_files["bam_reassigned_tmp"],
        )

    logging.info("BAM index not found. Indexing...")
    save = pysam.set_verbosity(0)
    samfile = pysam.AlignmentFile(out_bam, "rb", threads=threads)
    chr_lengths = []
    for chrom in samfile.references:
        chr_lengths.append(samfile.get_reference_length(chrom))
    max_chr_length = np.max(chr_lengths)
    pysam.set_verbosity(save)
    samfile.close()

    if max_chr_length > 536870912:
        logging.info("A reference is longer than 2^29, indexing with csi")
        pysam.index(
            "-c",
            "-@",
            str(threads),
            out_bam,
        )
    else:
        pysam.index(
            "-@",
            str(threads),
            out_bam,
        )

    os.remove(out_files["bam_reassigned_tmp"])


def calculate_alignment_score(identity, read_length):
    return (identity / math.log(read_length)) * math.sqrt(read_length)


def get_bam_data(parms, ref_lengths=None, percid=90, min_read_length=30, threads=1):
    bam, references = parms
    samfile = pysam.AlignmentFile(bam, "rb", threads=threads)

    results = []
    reads = set()
    refs = set()
    empty_df = 0

    for reference in references:
        if ref_lengths is None:
            reference_length = int(samfile.get_reference_length(reference))
        else:
            reference_length = int(ref_lengths[reference])
        aln_data = []
        for aln in samfile.fetch(
            contig=reference, multiple_iterators=False, until_eof=True
        ):
            # AS_value = aln.get_tag("AS") if aln.has_tag("AS") else None
            # XS_value = aln.get_tag("XS") if aln.has_tag("XS") else None
            query_length = (
                aln.query_length if aln.query_length != 0 else aln.infer_query_length()
            )

            if query_length < min_read_length:
                continue

            pident = (1 - ((aln.get_tag("NM") / query_length))) * 100
            if pident >= percid:
                var = calculate_alignment_score(pident, query_length)
                aln_data.append(
                    (
                        aln.query_name,
                        aln.reference_name,
                        var,
                        reference_length,
                    )
                )
                reads.add(aln.query_name)
                refs.add(aln.reference_name)
        # remove duplicates
        # check if aln_data is not empty
        if len(aln_data) > 0:
            aln_data_dt = dt.Frame(aln_data)
            # "queryId", "subjectId", "bitScore", "slen"
            aln_data_dt.names = ["queryId", "subjectId", "var", "slen"]
            # remove duplicates and keep the ones with the largest mapq
            aln_data_dt = aln_data_dt[
                :1, :, dt.by(dt.f.queryId, dt.f.subjectId), dt.sort(-dt.f.var)
            ]
            # aln_data_dt["slen"] = dt.Frame([ref_lengths[x] for x in aln_data_dt[:,"subjectId"].to_list()[0]])
            results.append(aln_data_dt)
        else:
            results.append(dt.Frame())
            empty_df += 1
    samfile.close()
    return (dt.rbind(results), reads, refs, empty_df)


def reassign_reads(
    bam,
    out_files,
    reference_lengths=None,
    threads=1,
    min_read_count=1,
    min_read_ani=90,
    min_read_length=30,
    reassign_iters=25,
    reassign_scale=0.9,
    sort_memory="4G",
    sort_by_name=False,
):
    dt.options.progress.enabled = True
    dt.options.progress.clear_on_success = True
    if threads > 1:
        dt.options.nthreads = threads - 1
    else:
        dt.options.nthreads = 1

    log.info("::: Loading BAM file")
    save = pysam.set_verbosity(0)
    samfile = pysam.AlignmentFile(bam, "rb", threads=threads)

    references = samfile.references

    pysam.set_verbosity(save)

    if reference_lengths is not None:
        ref_len_dt = dt.fread(reference_lengths)
        ref_len_dt.names = ["subjectId", "slen"]
        # convert to dict
        ref_len_dict = dict(
            zip(ref_len_dt["subjectId"].to_list()[0], ref_len_dt["slen"].to_list()[0])
        )
        # check if the dataframe contains all the References in the BAM file
        if not set(references).issubset(set(ref_len_dict.keys())):
            logging.error(
                "The BAM file contains references not found in the reference lengths file"
            )
            sys.exit(1)
    else:
        ref_len_dict = None

    total_refs = samfile.nreferences
    log.info(f"::: Found {total_refs:,} reference sequences")
    # logging.info(f"Found {samfile.mapped:,} alignments")
    log.info(f"::: Removing references with less than {min_read_count} reads...")
    references_m = {
        chrom.contig: chrom.mapped
        for chrom in samfile.get_index_statistics()
        if chrom.mapped >= min_read_count
    }

    references = list(references_m.keys())

    if len(references) == 0:
        log.warning("::: No reference sequences with alignments found in the BAM file")
        create_empty_output_files(out_files)
        sys.exit(0)

    log.info(f"::: Keeping {len(references):,} references")

    log.info("::: Creating reference chunks with uniform read amounts...")
    # ify the number of chunks
    ref_chunks = sort_keys_by_approx_weight(
        input_dict=references_m, scale=1, num_cores=threads
    )
    log.info(f"::: Created {len(ref_chunks):,} chunks")
    ref_chunks = random.sample(ref_chunks, len(ref_chunks))

    dt.options.progress.enabled = False
    dt.options.progress.clear_on_success = True
    dt.options.nthreads = 1

    parms = list(zip([bam] * len(ref_chunks), ref_chunks))

    log.info("::: Extracting reads from BAM file...")
    if is_debug():
        data = list(
            tqdm.tqdm(
                map(
                    partial(
                        get_bam_data,
                        ref_lengths=ref_len_dict,
                        percid=min_read_ani,
                        min_read_length=min_read_length,
                        threads=4,
                    ),
                    parms,
                    chunksize=1,
                ),
                total=len(parms),
                leave=False,
                ncols=80,
                desc="Chunks processed",
            )
        )
    else:
        p = Pool(
            threads,
        )
        data = list(
            tqdm.tqdm(
                p.imap_unordered(
                    partial(
                        get_bam_data,
                        ref_lengths=ref_len_dict,
                        percid=min_read_ani,
                        threads=4,
                    ),
                    parms,
                    chunksize=1,
                ),
                total=len(parms),
                leave=False,
                ncols=80,
                desc="Chunks processed",
            )
        )

        p.close()
        p.join()

    dt.options.progress.enabled = True
    dt.options.progress.clear_on_success = True
    if threads > 1:
        dt.options.nthreads = threads - 1
    else:
        dt.options.nthreads = 1

    log.info("::: Collecting results...")
    reads = list()
    refs = list()
    empty_df = 0
    for i in tqdm.tqdm(range(len(data)), total=len(data), leave=False, ncols=80):
        empty_df += data[i][3]
        reads.extend(list(data[i][1]))
        refs.extend(list(data[i][2]))
        data[i] = data[i][0]
    log.info(f"::: ::: Found {empty_df:,} references without alignments")

    data = dt.rbind([x for x in data])

    log.info("::: Indexing references...")
    refs = dt.Frame(list(set(refs)))
    refs.names = ["subjectId"]
    refs["sidx"] = dt.Frame(list(range(refs.shape[0])))
    refs.key = "subjectId"

    log.info("::: Indexing reads...")
    reads = dt.Frame(list(set(reads)))
    reads.names = ["queryId"]
    reads["qidx"] = dt.Frame([(i + refs.shape[0]) for i in range(reads.shape[0])])
    reads.key = "queryId"

    log.info("::: Combining data...")
    data = data[:, :, dt.join(reads)]
    data = data[:, :, dt.join(refs)]

    del data["queryId"]
    del data["subjectId"]
    n_alns_0 = data.shape[0]
    n_reads_0 = reads.shape[0]
    n_refs_0 = refs.shape[0]
    log.info(
        f"::: References: {n_refs_0:,} | Reads: {n_reads_0:,} | Alignments: {n_alns_0:,}"
    )

    log.info("::: Allocating data structures...")
    mat = data[:, [dt.f.qidx, dt.f.sidx, dt.f.var, dt.f.slen]].to_numpy()
    data = None

    # Create a zeros array with the same number of rows as 'mat'
    zeros_array = np.zeros((mat.shape[0], 5))

    # Stack the zeros_array with the original 'mat'
    m = np.column_stack([mat, zeros_array])
    zeros_array = None

    dtype = np.dtype(
        [
            ("source", "int"),
            ("subject", "int"),
            ("var", "float"),
            ("slen", "int"),
            ("s_W", "float"),
            ("prob", "float"),
            ("iter", "int"),
            ("n_aln", "int"),
            ("max_prob", "float"),
        ]
    )

    # Convert the unstructured array to structured array
    m = rf.unstructured_to_structured(m, dtype)
    gc.collect()
    log.info("::: Initializing data structures...")
    init_data = initialize_subject_weights(m)
    if reassign_iters > 0:
        log.info(f"::: Reassigning reads with {reassign_iters} iterations")
    else:
        log.info("::: Reassigning reads until convergence")
    no_multimaps = resolve_multimaps(
        init_data, iters=reassign_iters, scale=reassign_scale
    )

    n_reads = len(list(set(no_multimaps["source"])))
    n_refs = len(list(set(no_multimaps["subject"])))
    n_alns = no_multimaps.shape[0]
    log.info(
        f"::: References: {n_refs:,} | Reads: {n_reads:,} | Alignments: {n_alns:,}"
    )
    log.info(
        f'::: Unique mapping reads: {no_multimaps[no_multimaps["n_aln"] == 1].shape[0]:,} | Multimapping reads: {no_multimaps[no_multimaps["n_aln"] > 1].shape[0]:,}'
    )

    # add this to the array
    # no_multimaps["n_aln"] = subject_counts_array[no_multimaps["subject"]]

    # log.info(f"::: Removing references with less than {min_read_count} reads...")
    # no_multimaps = no_multimaps[no_multimaps["n_aln"] >= min_read_count]
    # log.info(f"{no_multimaps.shape[0]:,} alignments left")
    log.info("::: Mapping back indices...")
    if threads > 1:
        dt.options.nthreads = threads - 1
    else:
        dt.options.nthreads = 1

    g = dt.Frame(no_multimaps["source"])
    g.names = ["qidx"]
    reads.key = "qidx"
    q = g[:, :, dt.join(reads)]

    g = dt.Frame(no_multimaps["subject"])
    g.names = ["sidx"]
    refs.key = "sidx"
    s = g[:, :, dt.join(refs)]

    log.info("::: Calculating reads per subject...")
    # count how many alignments are in each subjectId
    s_c = s[:, dt.count(dt.f.subjectId), dt.by(dt.f.subjectId)]
    s_c.names = ["subjectId", "counts"]
    references_m = dict()
    log.info(f"::: Removing references with less than {min_read_count:,}...")
    for i, k in zip(s_c[:, "subjectId"].to_list()[0], s_c[:, "counts"].to_list()[0]):
        if k >= min_read_count:
            references_m[i] = k
    log.info(f"::: ::: Keeping {len(references_m):,} references")
    s_c = None
    if len(references_m) == 0:
        log.warning("::: No reference sequences with alignments found in the BAM file")
        create_empty_output_files(out_files)
        sys.exit(0)
    # convert columns queryId from q and subjectId from s to a tuple
    log.info("::: Creating filtered set...")
    entries = defaultdict(set)
    q_query_ids = q[:, "queryId"].to_list()[0]
    s_subject_ids = s[:, "subjectId"].to_list()[0]

    for query_id, subject_id in zip(q_query_ids, s_subject_ids):
        if subject_id in references_m:
            entries[subject_id].add((query_id, subject_id))
    no_multimaps = None
    q = None
    s = None
    q_query_ids = None
    s_subject_ids = None
    gc.collect()
    log.info("::: Writing to BAM file...")
    write_reassigned_bam(
        bam=bam,
        ref_counts=references_m,
        out_files=out_files,
        threads=threads,
        entries=entries,
        sort_memory=sort_memory,
        min_read_ani=min_read_ani,
    )


def reassign(args):
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(levelname)s ::: %(asctime)s ::: %(message)s",
        datefmt="%H:%M:%S",
    )

    args = get_arguments()
    bam = args.bam
    tmp_dir = check_tmp_dir_exists(args.tmp_dir)
    log.info("Temporary directory: %s", tmp_dir.name)
    out_files = create_output_files(
        prefix=args.prefix,
        bam=args.bam,
        tmp_dir=tmp_dir,
        mode="reassign",
        # bam_reassigned=args.bam_reassigned,
    )

    bam = check_bam_file(
        bam=args.bam,
        threads=args.threads,
        reference_lengths=args.reference_lengths,
        sort_memory=args.sort_memory,
    )
    if bam is None:
        logging.warning("No reference sequences with alignments found in the BAM file")
        create_empty_output_files(out_files)
        sys.exit(0)

    logging.getLogger("my_logger").setLevel(
        logging.DEBUG if args.debug else logging.INFO
    )

    if args.debug:
        warnings.showwarning = handle_warning
    else:
        warnings.filterwarnings("ignore")
    logging.info("Resolving multi-mapping reads...")
    reassign_reads(
        bam=bam,
        threads=args.threads,
        reference_lengths=args.reference_lengths,
        min_read_count=args.min_read_count,
        min_read_ani=args.min_read_ani,
        min_read_length=args.min_read_length,
        reassign_iters=args.reassign_iters,
        reassign_scale=args.reassign_scale,
        sort_memory=args.sort_memory,
        sort_by_name=args.sort_by_name,
        out_files=out_files,
    )
    log.info("Done!")



import pysam
import taxopy as txp
from tqdm import tqdm
from multiprocessing import Pool, Manager
from functools import partial
import pandas as pd
import logging
import networkx as nx
import sys
from bam_filter.utils import (
    calc_chunksize,
    sort_keys_by_approx_weight,
    concat_df,
    is_debug,
    create_output_files,
    create_empty_output_files,
)
from bam_filter.sam_utils import check_bam_file
from collections import defaultdict
from functools import reduce
import operator
import random
import gzip

log = logging.getLogger("my_logger")

debug = is_debug()


def calculate_path_likelihood(path, graph):
    likelihood = 1.0
    for u, v in zip(path[:-1], path[1:]):
        likelihood *= graph[u][v]["cum_weight"]
    return likelihood


def find_most_likely_continuation_worker(partial_path_end, full_graph, index):
    # find where are we in the taxonomy tree, if we are in the root, return

    results = []
    try:
        descendants = list(nx.descendants(full_graph, partial_path_end))
    except nx.NetworkXError:
        descendants = []

    if not descendants:
        return (
            partial_path_end,
            {
                "reference": None,
                "best_path": None,
                "top_10_paths": None,
            },
        )

    if (
        len(nx.shortest_path(full_graph, source="root", target=partial_path_end))
        <= index + 1
    ):
        return (
            partial_path_end,
            {
                "reference": None,
                "best_path": None,
                "top_10_paths": None,
            },
        )

    for continuation in descendants:
        full_path = list(
            nx.all_simple_paths(
                full_graph, source=partial_path_end, target=continuation
            )
        )[0]
        likelihood = calculate_path_likelihood(full_path, full_graph)
        results.append((full_path, likelihood))

    results.sort(key=lambda x: x[1], reverse=True)

    return (
        partial_path_end,
        {
            "reference": results[0][0][-1] if results else None,
            "best_path": results[0] if results else None,
            "top_10_paths": results[:10] if results else None,
        },
    )


def find_most_likely_continuation(full_graph, leaves, index):
    result_dict = {}

    for partial_path_end in tqdm(leaves, leave=False, ncols=80):
        partial_path_end, result = find_most_likely_continuation_worker(
            partial_path_end, full_graph, index
        )
        result_dict[partial_path_end] = result

    return result_dict


def calculate_cumulative_weight_and_path(graph, node, lengths):
    # Base case: if the node is the root, return its cumulative weight and path
    if not list(graph.predecessors(node)):
        if node in lengths:
            return 0, 0, [node]
        else:
            # nei = list(graph.neighbors(node))[0]
            # cumulative_weight = graph[node][nei].get("weight", 0)
            # cumulative_norm_weight = graph[node][nei].get("norm_weight", 0)
            # return cumulative_weight, cumulative_norm_weight, [node]
            return 0, 0, [node]

    # Recursive case: calculate cumulative weight and path by summing the edge weight
    # and norm_weight with the cumulative weight and path of its parent(s)
    cumulative_weight = 0
    cumulative_norm_weight = 0
    cumulative_path = []

    for parent in graph.predecessors(node):
        edge_weight = graph[parent][node].get("weight", 0)
        norm_weight = graph[parent][node].get("norm_weight", 0)
        (
            parent_weight,
            parent_norm_weight,
            parent_path,
        ) = calculate_cumulative_weight_and_path(graph, parent, lengths)

        cumulative_weight += edge_weight + parent_weight
        cumulative_norm_weight += norm_weight + parent_norm_weight
        cumulative_path.extend(parent_path + [node])  # Fix: Use += to concatenate lists

    return cumulative_weight, cumulative_norm_weight, cumulative_path


def create_tax_graph_w(tax_path, weight, lengths, scale=1_000_000):
    root_row = pd.DataFrame({"source": ["root"], "target": ["root"], "weight": 0})
    res = list(zip(tax_path, tax_path[1:]))
    # get last element
    res = pd.DataFrame(res, columns=["source", "target"])
    res = pd.concat([root_row, res])
    res = res.drop_duplicates()
    res["weight"] = 0
    res["norm_weight"] = 0
    res.iloc[-1, res.columns.get_loc("weight")] = weight
    res.iloc[-1, res.columns.get_loc("norm_weight")] = round(
        (weight / lengths[res.iloc[-1, res.columns.get_loc("target")]]) * scale
    )
    return res


def create_lca_df(tax_path, weight):
    root_row = pd.DataFrame({"source": ["root"], "target": ["root"], "weight": 0})
    res = list(zip(tax_path, tax_path[1:]))
    # get last element
    res = pd.DataFrame(res, columns=["source", "target"])
    res = pd.concat([root_row, res])
    res = res.drop_duplicates()
    res["weight"] = 0
    # add 1 to the last row weight
    res.iloc[-1, res.columns.get_loc("weight")] = weight
    return res


def get_ref2read(params, dat, threads=1):
    bam, references = params
    samfile = pysam.AlignmentFile(bam, "rb", threads=threads)
    results = defaultdict(set)
    for reference in references:
        if reference not in dat:
            continue
        for aln in samfile.fetch(
            contig=reference, multiple_iterators=False, until_eof=True
        ):
            results[reference].add(aln.query_name)
    samfile.close()
    return results


def get_tax(ref, parms):
    taxdb = parms["taxdb"]
    acc2taxid = parms["acc2taxid"]
    if ref in acc2taxid:
        taxid = acc2taxid[ref]
        # taxid = txp.taxid_from_name(ref, taxdb)[0]
        taxonomy_info = txp.Taxon(taxid, taxdb).rank_name_dictionary
        taxonomy_info["taxid"] = taxid
        taxonomy_info["ref"] = ref
        taxonomy_info["subspecies"] = f"S__{ref}"
    else:
        log.debug(f"No taxid found for {ref}")
        taxonomy_info = None
    return taxonomy_info


def get_taxonomy_info(refids, taxdb, acc2taxid, nprocs=1):
    """Function to get the references taxonomic information for a given taxonomy id

    Args:
        taxids (list): A list of taxonomy ids
        taxdb (taxopy.TaxonomyDB): A taxopy DB

    Returns:
        dict: A list of taxonomy information
    """

    acc2taxid_df = pd.read_csv(acc2taxid, sep="\t", index_col=None)[
        ["accession", "taxid"]
    ].rename(columns={"accession": "reference"}, inplace=False)
    # Filter rows in refids from dataframe
    acc2taxid_df = acc2taxid_df.loc[acc2taxid_df["reference"].isin(refids)]
    acc2taxid_dict = acc2taxid_df.set_index("reference").T.to_dict("records")

    parms = {"taxdb": taxdb, "acc2taxid": acc2taxid_dict[0]}
    func = partial(get_tax, parms=parms)
    if debug is True or len(refids) < 100000:
        taxonomy_info = list(map(func, refids))
    else:
        p = Pool(nprocs)
        c_size = calc_chunksize(nprocs, len(refids))
        taxonomy_info = list(
            tqdm(
                p.imap_unordered(func, refids, chunksize=c_size),
                total=len(refids),
                leave=False,
                ncols=100,
                desc="References processed",
            )
        )
        p.close()
        p.join()
    taxonomy_info = list(filter(None, taxonomy_info))
    exclude = ["taxid", "ref"]
    tax_ranks = []

    for k in taxonomy_info[0].keys():
        if k not in exclude:
            tax_ranks.append(k)

    taxonomy_info = {i["ref"]: i for i in taxonomy_info}
    return taxonomy_info, tax_ranks


def do_lca(args):
    bam = args.bam
    names = args.names
    nodes = args.nodes
    acc2taxid = args.acc2taxid
    sel_rank = args.rank_lca
    reference_lengths = args.reference_lengths
    threads = args.threads
    scale = args.scale

    out_files = create_output_files(
        bam=bam,
        prefix=args.prefix,
        lca_summary=args.lca_summary,
        tmp_dir=None,
        mode="lca",
    )

    bam = check_bam_file(
        bam=args.bam,
        threads=args.threads,
        reference_lengths=args.reference_lengths,
        sort_memory=args.sort_memory,
    )
    if bam is None:
        logging.warning("No reference sequences with alignments found in the BAM file")
        create_empty_output_files(out_files)
        sys.exit(1)

    samfile = pysam.AlignmentFile(bam, "rb", threads=threads)
    references = samfile.references
    references_m = {
        chrom.contig: chrom.mapped for chrom in samfile.get_index_statistics()
    }

    if reference_lengths is not None:
        ref_lengths = pd.read_csv(
            reference_lengths, sep="\t", index_col=False, names=["reference", "length"]
        )
        ref_lengths = dict(zip(ref_lengths["reference"], ref_lengths["length"]))
        # check if the dataframe contains all the References in the BAM file
        if not set(references).issubset(set(ref_lengths.keys())):
            logging.error(
                "The BAM file contains references not found in the reference lengths file"
            )
            sys.exit(1)
    else:
        ref_lengths = {x: samfile.get_reference_length(x) for x in references}

    log.info("Getting taxonomy information")
    taxdb = txp.TaxDb(
        nodes_dmp=nodes,
        names_dmp=names,
    )
    acc2taxid = acc2taxid
    taxonomy_info, tax_ranks = get_taxonomy_info(
        references, taxdb, acc2taxid, nprocs=threads
    )
    tax_ranks.reverse()
    index_r = tax_ranks.index(sel_rank)

    dat = {}
    for k, v in taxonomy_info.items():
        try:
            tax_list = ["root"]
            tax_list.extend([v[rank] for rank in tax_ranks])
            tax_list.append(k)
            dat[k] = tuple(tax_list)
        except KeyError:
            continue

    log.info("Getting reference to read mapping")
    ref_chunks = sort_keys_by_approx_weight(
        references_m, scale=1, num_cores=threads, refinement_steps=100
    )

    ref_chunks = random.sample(ref_chunks, len(ref_chunks))
    dat_man = Manager().dict(dat)
    params = zip([bam] * len(ref_chunks), ref_chunks)
    p = Pool(
        threads,
    )
    data = list(
        tqdm(
            p.imap_unordered(
                partial(
                    get_ref2read,
                    dat=dat_man,
                    threads=1,
                ),
                params,
                chunksize=1,
            ),
            total=len(ref_chunks),
            leave=False,
            ncols=80,
            desc="Chunks processed",
        )
    )

    p.close()
    p.join()

    data = reduce(operator.ior, data, {})

    log.info("Getting reads taxonomic information")
    reads = defaultdict(set)
    for k, v in tqdm(data.items(), total=len(data), leave=False, ncols=80):
        tax = dat[k]
        for read in v:
            reads[read].add(tax)

    log.info("Creating taxonomic graph")
    root_row = pd.DataFrame({"source": ["root"], "target": ["root"], "weight": [0]})
    gs = []
    if len(dat) > 0:
        for t in tqdm(dat.values(), total=len(dat), leave=False, ncols=80):
            # remove last element
            # t = t[:-1]
            res = list(zip(t, t[1:]))
            res = pd.DataFrame(res, columns=["source", "target"])
            res = res.drop_duplicates()
            res["weight"] = 0
            res = pd.concat([root_row, res])
            gs.append(res)

    gs = concat_df(gs).drop_duplicates()

    G = nx.from_pandas_edgelist(gs, create_using=nx.DiGraph(), edge_attr=True)

    dat = None

    log.info("Calculating LCA")
    cum_tax = defaultdict(int)
    unique = []

    discarded_lca = defaultdict(int)
    for_lca = defaultdict(int)

    # for k, v in tqdm(reads.items(), total=len(reads), leave=False, ncols=80):
    #     v = list(v)
    #     if len(v) > 1:
    #         tax_path = [edge for edge in v[0] if all(edge in t for t in v[1:])]
    #         if len(tax_path) <= index_r + 1:
    #             discarded_lca[tuple(tax_path)] += 1
    #             continue
    #         for_lca[tuple(tax_path)] += 1
    #     else:
    #         unique.append(k)
    #         cum_tax[v[0]] += 1

    for k, v in tqdm(reads.items(), total=len(reads), leave=False, ncols=80):
        v = list(v)
        if len(v) > 1:
            tax_path_set = set(v[0])
            for t in v[1:]:
                tax_path_set.intersection_update(t)

            tax_path = [edge for edge in v[0] if edge in tax_path_set]

            if len(tax_path) <= index_r + 1:
                discarded_lca[tuple(tax_path)] += 1
                continue

            for_lca[tuple(tax_path)] += 1
        else:
            if len(k) <= index_r + 1:
                discarded_lca[tuple(k)] += 1
                continue
            unique.append(k)
            cum_tax[v[0]] += 1

    log.info(
        f"Unique: {len(unique)} | LCA: {sum(for_lca.values())} | Discarded: {sum(discarded_lca.values())}"
    )

    log.info("Creating unique mapping taxonomic graph")
    gs = []
    for k, v in tqdm(cum_tax.items(), total=len(cum_tax), leave=False, ncols=80):
        gs.append(create_tax_graph_w(list(k), v, ref_lengths, scale=scale))

    df = concat_df(gs)

    if len(for_lca) > 0:
        log.info("Creating taxonomic graph with LCA nodes")
        for_lca_df = []
        for k, v in tqdm(for_lca.items(), total=len(for_lca), leave=False, ncols=80):
            for_lca_df.append(create_lca_df(list(k), v))
        df_l = concat_df(for_lca_df)
        df_l["weight"] = 0
        df_l["norm_weight"] = 0
        df = concat_df([df, df_l])
    # group by source and target and sum weight
    df = df.groupby(["source", "target"]).sum().reset_index()

    G = nx.from_pandas_edgelist(df, create_using=nx.DiGraph(), edge_attr=True)
    G.remove_edges_from(nx.selfloop_edges(G))
    Gr = G.reverse()

    log.info("Calculating cumulative weights and paths on the taxonomic graph")
    cumulative_weights_and_paths = {}
    modified_graph = Gr.copy()  # Create a copy of the original graph
    for node in Gr.nodes:
        (
            cumulative_weight,
            cumulative_norm_weight,
            path,
        ) = calculate_cumulative_weight_and_path(modified_graph, node, ref_lengths)
        cumulative_weights_and_paths[node] = {
            "cumulative_weight": cumulative_weight,
            "cumulative_norm_weight": cumulative_norm_weight,
            "path": ";".join(nx.shortest_path(G, target=node)["root"]),
        }

        # Update the modified graph with inferred edge attributes after all weights have been calculated
        for parent in modified_graph.predecessors(node):
            modified_graph[parent][node]["cum_weight"] = (
                modified_graph[parent][node].get("weight", 0) + cumulative_weight
            )
            modified_graph[parent][node]["cum_norm_weight"] = (
                modified_graph[parent][node].get("norm_weight", 0)
                + cumulative_norm_weight
            )

    cumulative_weights_and_paths = dict(
        sorted(cumulative_weights_and_paths.items(), key=lambda item: item[1]["path"])
    )

    if len(for_lca) > 0:
        df1 = concat_df(for_lca_df)
        df1 = df1.groupby(["source", "target"]).sum().reset_index()
        G_lca = nx.from_pandas_edgelist(df1, create_using=nx.DiGraph(), edge_attr=True)
        G_lca.remove_edges_from(nx.selfloop_edges(G_lca))

        df1_d = dict(
            zip(df1[df1["weight"] > 0]["target"], df1[df1["weight"] > 0]["weight"])
        )
        leaves = list(df1_d.keys())

        log.info(f"Finding most likely reference for the LCA nodes")
        lca2ref = find_most_likely_continuation(
            full_graph=modified_graph.reverse(),
            leaves=leaves,
            index=index_r,
        )

        lca_dfs = []
        for k, v in lca2ref.items():
            tax_path = nx.shortest_path(G, target=k)["root"]

            root_row = pd.DataFrame(
                {"source": ["root"], "target": ["root"], "weight": 0}
            )
            res = list(zip(tax_path, tax_path[1:]))
            # get last element
            res = pd.DataFrame(res, columns=["source", "target"])
            res = pd.concat([root_row, res])
            res = res.drop_duplicates()
            res["weight"] = 0
            res["norm_weight"] = 0
            res.iloc[-1, res.columns.get_loc("weight")] = df1_d[k]
            if v["reference"] is None:
                res.iloc[-1, res.columns.get_loc("norm_weight")] = df1_d[k]
            else:
                res.iloc[-1, res.columns.get_loc("norm_weight")] = round(
                    scale * df1_d[k] / ref_lengths[v["reference"]]
                )
            lca_dfs.append(res)
        log.info("Adding LCA nodes to the taxonomic graph")
        df2 = concat_df(lca_dfs)
        df2 = concat_df([df2, df])
        df2 = df2.groupby(["source", "target"]).sum().reset_index()
        # %%
        G2 = nx.from_pandas_edgelist(df2, create_using=nx.DiGraph(), edge_attr=True)
        G2.remove_edges_from(nx.selfloop_edges(G2))
        Gr2 = G2.reverse()
        # %%
        log.info("Calculating cumulative weights and paths on the taxonomic graph")
        cumulative_weights_and_paths = {}
        modified_graph = Gr2.copy()  # Create a copy of the original graph

        for node in Gr2.nodes:
            (
                cumulative_weight,
                cumulative_norm_weight,
                path,
            ) = calculate_cumulative_weight_and_path(modified_graph, node, ref_lengths)

            cumulative_weights_and_paths[node] = {
                "cumulative_weight": cumulative_weight,
                "cumulative_norm_weight": cumulative_norm_weight,
                "path": ";".join(nx.shortest_path(G, target=node)["root"]),
            }

            # Update the modified graph with inferred edge attributes after all weights have been calculated
            for parent in modified_graph.predecessors(node):
                modified_graph[parent][node]["cum_weight"] = (
                    modified_graph[parent][node].get("weight", 0) + cumulative_weight
                )
                modified_graph[parent][node]["cum_norm_weight"] = (
                    modified_graph[parent][node].get("norm_weight", 0)
                    + cumulative_norm_weight
                )
        cumulative_weights_and_paths = dict(
            sorted(
                cumulative_weights_and_paths.items(), key=lambda item: item[1]["path"]
            )
        )

    log.info("Writing LCA results to file")
    nodes = list(
        set(
            [
                node
                for node, data in cumulative_weights_and_paths.items()
                if data["cumulative_weight"] > 0
            ]
        )
    )
    taxids = txp.taxid_from_name(nodes, taxdb)
    taxids = dict(zip(nodes, taxids))

    out_file = out_files["lca_summary"]
    # determine if the file is gzipped or not
    if out_file.endswith(".gz"):
        with gzip.open(out_file, "wt") as f:
            f.write("taxid\tname\trank\tn_reads\tabundance\ttax_path\n")
            for node, data in tqdm(
                cumulative_weights_and_paths.items(),
                total=len(cumulative_weights_and_paths),
                leave=False,
                ncols=80,
                desc="Writing results",
                unit="nodes",
            ):
                if data["cumulative_weight"] > 0:
                    taxid = taxids[node][0]
                    rank = taxdb.taxid2rank[taxid]
                    f.write(
                        f"{taxid}\t{node}\t{rank}\t{data['cumulative_weight']}\t{data['cumulative_norm_weight']}\t{data['path']}\n"
                    )
    else:
        with open(out_file, "wt") as f:
            f.write("taxid\tname\trank\tn_reads\tabundance\ttax_path\n")
            for node, data in tqdm(
                cumulative_weights_and_paths.items(),
                total=len(cumulative_weights_and_paths),
            ):
                if data["cumulative_weight"] > 0:
                    taxid = taxids[node][0]
                    rank = taxdb.taxid2rank[taxid]
                    f.write(
                        f"{taxid}\t{node}\t{rank}\t{data['cumulative_weight']}\t{data['cumulative_norm_weight']}\t{data['path']}\n"
                    )
